/* -
 *
 * Author: Damien Masson
 * Copyright University of Lille / Inria
 *
 * http://ns.inria.fr/mjolnir/whichfingers/
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 or any later version.
 *
 */


#include <SPI.h>
#include <WiFi101.h>
#include <WiFiUdp.h>
#include "WhichFingers.h" // To send data through HID (only used for DEBUG)
#define LEFT_HAND 0
#define RIGHT_HAND 1

/***********
   Configuration
*/
#define HAND_SIDE RIGHT_HAND // Specify on which hand this prototype will be equipped

#define BUFFER_SIZE 500
#define NB_PIEZO 5
#define FIRST_PIEZO A2
#define DELAY 990 // Delay in microseconds between two reads, 993 gives approximately 1000Hz
#define DATA_WINDOW_LEFT 10
#define DATA_WINDOW_RIGHT 10

#define DEFAULT_TIME_WINDOW_LEFT 20
#define DEFAULT_TIME_WINDOW_RIGHT 0

#define DEBUG 1 // If DEBUG is enabled, more info are outputed + piezo values are sent through HID too
#define USE_SYNC 0

#define USE_WIFI 1

char ssid[] = "TestWF";
char pass[] = "test1234";

IPAddress ip(192, 168, 1, 43); // Default server IP, but it should never be used as the ip is found automatically
IPAddress broadcastIp(255, 255, 255, 255);
int port = 8770; // Port used for the communication (UDP and TCP)



/*
   Structures of the messages sent & received
*/
typedef struct _finger_request {
  unsigned char id;
  unsigned char is_calibration;
} finger_request;

typedef struct _finger_response {
  unsigned char maxValues[5];
  unsigned char id;
} finger_response;

typedef struct _handshake_packet {
  unsigned char handSide;
} handshake_packet;


unsigned char buff[NB_PIEZO][BUFFER_SIZE];
int idx;

WiFiUDP* udp = NULL;
WiFiClient* client = NULL;

int status = WL_IDLE_STATUS;

int time_window_left = DEFAULT_TIME_WINDOW_LEFT;
int time_window_right = DEFAULT_TIME_WINDOW_RIGHT;
unsigned char calibration_peaks[500] = {0};
unsigned int nb_calibration_values = 0;

/*
   Wait until connected to the Wifi network specified in the configuration
*/
void connectToWifi() {
  if (WiFi.status() == WL_NO_SHIELD) {
    Serial.println("WiFi shield not present");
    while (true);
  }

  while (status != WL_CONNECTED) {
    Serial.print("Attempting to connect to SSID: ");
    Serial.println(ssid);
    status = WiFi.begin(ssid, pass);
    delay(1000);
  }
  Serial.println("Connected to wifi");
}

/*
   Wait until connected to the TCP server specified in the configuration
*/
void connectToServer() {
  // WiFiUDP and WiFiClient can't cohabit (otherwise, the program freeze), so we have to keep only one at the same time
  if (client != NULL) {
    client->stop();
    delete client;
    client = NULL;
  }

  // Find the ip of the server by broadcasting a handshake message on the network
  Serial.println("Searching the ip address of the server...");

  handshake_packet handshake;
  handshake.handSide = HAND_SIDE;

  udp = new WiFiUDP;

  udp->begin(port);
  do {
    udp->beginPacket(broadcastIp, port);
    udp->write((uint8_t*) &handshake, sizeof(handshake_packet));
    udp->endPacket();
    Serial.println("Broadcasting handshake to find a server...");
    delay(1000);
    udp->parsePacket();
    int sizeAvailable = udp->available(); 
    if (sizeAvailable == sizeof(handshake_packet)) {
      handshake_packet response;
      udp->read((uint8_t*) &response, sizeof(handshake_packet));
      if (response.handSide == HAND_SIDE) {
        break;
      }
    }
  } while (true);
  ip = udp->remoteIP();
  udp->stop();
  delete udp;

  client = new WiFiClient;
  do {
    Serial.print("Attempting to connect to server: ");
    Serial.print(ip);
    Serial.print(" on port: ");
    Serial.println(port);
    client->connect(ip, port);
    delay(1000);
  } while (!client->connected());
  Serial.println("Connected to server");
}

/*
   Boost the ADC to read analog pins faster
   Needed to read the 5 piezos at 1000Hz on the MKR1000
*/
void AdcBooster()
{
  ADC->CTRLA.bit.ENABLE = 0;                     // Disable ADC
  while ( ADC->STATUS.bit.SYNCBUSY == 1 );       // Wait for synchronization
  ADC->CTRLB.reg = ADC_CTRLB_PRESCALER_DIV32 |   // Divide Clock by 64.
                   ADC_CTRLB_RESSEL_10BIT |      // Result on 10 bits
                   ADC_CTRLB_CORREN;
  ADC->OFFSETCORR.reg = 0;
  ADC->GAINCORR.reg = 0b100000000000;

  ADC->AVGCTRL.reg = ADC_AVGCTRL_SAMPLENUM_4 |   // 1 sample
                     ADC_AVGCTRL_ADJRES(0x02ul); // Adjusting result by 0
  //ADC->CTRLB.bit.FREERUN = 0x1;
  ADC->SAMPCTRL.reg = 0x00;                      // Sampling Time Length = 0
  ADC->CTRLA.bit.ENABLE = 1;                     // Enable ADC
  while ( ADC->STATUS.bit.SYNCBUSY == 1 );       // Wait for synchronization
}


void setup() {
  Serial.begin(9600);
  #if USE_WIFI
  connectToWifi();
  connectToServer();
  #endif
  AdcBooster();

#if DEBUG
  WhichFingers.begin();
#endif
}

/*
   Get the value of a piezo
*/
uint8_t last_spike_idx = A2;
uint8_t last_spike_time = 0;
uint8_t getPiezoValue(int idx) {
  // On the right hand, the pins are inverted (because the board is inverted)
#if HAND_SIDE == RIGHT_HAND
  idx = (NB_PIEZO - 1) - (idx - FIRST_PIEZO) + FIRST_PIEZO;
#endif


  int v = 0;
  int counter = 0;
  while (v == 0 && counter < 40) {
    v = analogRead(idx);
    ++counter;
  }

  if (v < 0) v = 0;
  if (v > 255) v = 255;


  ++last_spike_time;

  if (v > 10) {
    last_spike_idx = idx;
    last_spike_time = 0;
  }
  return (uint8_t) v;
}

/*
   Get the maximum values each piezo has reached during a specific interval (specified by the pos + configuration)
*/
void getPiezoMaxValues(unsigned int pos, unsigned char* maxValues) {

  for (int id_piezo = 0; id_piezo < NB_PIEZO; ++id_piezo) {
    uint8_t max_val = 0;
    uint8_t previous_val = 0;
    uint8_t previous_max_val = 0;
    bool suspicious = false;

    for (int i = time_window_left; i >= time_window_right; --i) {
      int idx = pos - i;
      if (idx < 0) idx += BUFFER_SIZE;
      uint8_t val = buff[id_piezo][idx];

#if DEBUG
      Serial.print(val);
      Serial.print(" ");
#endif

      // Sometimes, the value read are wrong, this code is to correct these values
      // These values are easily spotted because the value read before and after are completely diffent
      if (suspicious && previous_val - val > 10) {
        max_val = previous_max_val;
        Serial.print("(i)");
      }

      suspicious = false;
      if (val > max_val && i != 0) {
        previous_max_val = max_val;
        max_val = val;
        suspicious = (val - previous_val > 10);
      }

      previous_val = val;
    }
    maxValues[id_piezo] = max_val;

#if DEBUG
    Serial.println();
#endif
  }
}

void calibrateOnFinger(unsigned int currPos, unsigned char fingerUsed) {
  if (fingerUsed == 255) {
    // Request to clear the current calibration
    nb_calibration_values = 0;
    time_window_left = DEFAULT_TIME_WINDOW_LEFT;
    time_window_right = DEFAULT_TIME_WINDOW_RIGHT;
    Serial.println("Calibration reset");
    return;
  }
  
  // Look through the last 100 values of the finger to find his highest peak
  unsigned char highestVal = 0;
  int peakPos = 0;
  for (int i = 0; i < 100; ++i) {
      int idx = currPos - i;
      if (idx < 0) idx += BUFFER_SIZE;
      if (buff[fingerUsed][idx] > 30) {
        calibration_peaks[nb_calibration_values % sizeof(calibration_peaks)] = i;
        ++nb_calibration_values;
      }
  }

  int sum = 0;
  int nbVal = 0;
  for (; nbVal < nb_calibration_values; ++nbVal) {
    if (nbVal >= sizeof(calibration_peaks)) break;
    sum += calibration_peaks[nbVal];
  }

  double avg = ((double) sum) / nbVal;

  Serial.print("Average : ");
  Serial.println(avg);

  double dev = 0;
  for (int j = 0; j < nbVal; ++j) {
    double delta = calibration_peaks[j] - avg;
    dev += delta * delta;
  }
  dev = sqrt(dev / nbVal);

  Serial.print("Standard deviation : ");
  Serial.println(dev);

  time_window_left = round(avg + dev);
  time_window_right = round(max(avg - dev, 0));

  Serial.print("New time window : t");
  Serial.print(time_window_left);
  Serial.print(" t");
  Serial.print(time_window_right);
  Serial.println();
}

int offTime = 0;
unsigned long lastTimestamp = 0;

void loop() {
  unsigned long startTime = micros();
  unsigned long timestamp = millis();
  unsigned int idx = timestamp % BUFFER_SIZE;

  int len = 0;
  #if USE_WIFI
  while ((len = client->available()) >= sizeof(finger_request)) {
    Serial.print("Available : ");
    Serial.println(len);
    Serial.print("read : ");
    Serial.println(sizeof(finger_request));
    finger_request request;
    client->read((uint8_t*) &request, sizeof(finger_request));
    if (request.is_calibration) {
      Serial.println("Calibration !");
      calibrateOnFinger(idx, request.id);
    } else {
      Serial.print("Request !");
      finger_response response;
      getPiezoMaxValues(timestamp % BUFFER_SIZE, response.maxValues);
      Serial.print("Available after read : ");
      Serial.println(client->available());
      response.id = request.id;
      client->write((uint8_t*) &response, sizeof(finger_response));
      client->flush();
    }
  }

  // Test if we're still connected. Using client->connected() is laggy so I found this other way by reading the source code of WifiClient
  if (!(client->_flag & SOCKET_BUFFER_FLAG_CONNECTED)) {
    Serial.println("Disconnected from server");
    connectToServer();
  }
  #endif

  unsigned int lostMs = timestamp - lastTimestamp;
  if (lostMs > 1 && lostMs < 20) {
    Serial.print("Lost ");
    Serial.print(lostMs);
    Serial.println("ms");
    for (int i = 0; i < lostMs; ++i) {
      unsigned int pos = idx - i;
      if (pos < 0) pos += BUFFER_SIZE;
      for (int i = FIRST_PIEZO; i < FIRST_PIEZO + NB_PIEZO; ++i) {
        buff[i - FIRST_PIEZO][pos] = 0;
      }
    }
  }

  uint8_t hid_buff[5];
  for (int i = FIRST_PIEZO; i < FIRST_PIEZO + NB_PIEZO; ++i) {
    uint8_t v = getPiezoValue(i);
    uint8_t vMiddle = buff[i - FIRST_PIEZO][(timestamp - 1) % BUFFER_SIZE];
    uint8_t vLeft = buff[i - FIRST_PIEZO][(timestamp - 2) % BUFFER_SIZE];

    if (lastTimestamp == timestamp) {
      if (buff[i - FIRST_PIEZO][idx] == 0) {
        buff[i - FIRST_PIEZO][idx] = v;
      }
    } else {
      buff[i - FIRST_PIEZO][idx] = v;
      hid_buff[i - FIRST_PIEZO] = v;
    }
  }

#if DEBUG
  // Send the values by HID too
  if (lastTimestamp != timestamp) {
    WhichFingers.send(HAND_SIDE + 1, hid_buff);
  }
#endif


  // Wait the appropriate number of ms and take into account lags
  lastTimestamp = timestamp;

  int elapsed = micros() - startTime;

  if (elapsed > DELAY) {
    offTime += elapsed - DELAY;
  }

  if (offTime <= 0) {
    delayMicroseconds(DELAY - elapsed);
  } else {
    if (elapsed < DELAY) {
      offTime -= DELAY - elapsed;
      if (offTime < 0) offTime = 0;
    }
  }
}
