/* -
 *
 * Author: Damien Masson
 * Copyright University of Lille / Inria
 *
 * http://ns.inria.fr/mjolnir/whichfingers/
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 or any later version.
 *
 */

#include <Qt>
#include <QTcpServer>
#include <QUdpSocket>
#include <QTcpSocket>
#include <QDebug>
#include <QMessageBox>
#include <QDateTime>
#include <QEventLoop>

#include "whichfingers.h"

WhichFingers::WhichFingers() :
    QObject(NULL),
    tcpServer(NULL),
    udpServer(NULL),
    udpSocket(NULL)
{
    tcpServer = new QTcpServer(this);
    udpServer = new QUdpSocket(this);
    udpSocket = new QUdpSocket(this);

    memset(handClient, 0, sizeof(handClient));
    curPacketId = 0;

    startTimestamp = QDateTime::currentMSecsSinceEpoch();
    connect(tcpServer, SIGNAL(newConnection()), this, SLOT(acceptConnection()));
    connect(udpServer, SIGNAL(readyRead()), this, SLOT(receiveUdpData()));

    // The UDPServer is only used to establish the connection between the prototype and the server. The prototype will broadcast a message and the receiver will respond to it
    // that way the prototype can find the ip of the server automatically
    if (!udpServer->bind(QHostAddress::Any, 8770)) {
        QMessageBox::critical(NULL, tr("WhichFingers Server"), tr("Unable to start the UDP server: %1.").arg(udpServer->errorString()));
    }

    // The TCPServer will receive the values of the piezos
    if (!tcpServer->listen(QHostAddress::Any, 8770)) {
        QMessageBox::critical(NULL, tr("WhichFingers Server"), tr("Unable to start the TCP server: %1.").arg(tcpServer->errorString()));
        return;
    }
}

/*
 * Accept a TCP connection
 */
void WhichFingers::acceptConnection() {
    QTcpSocket* client = tcpServer->nextPendingConnection();

    unsigned char handSide = hostHandSide[client->peerAddress()];
    connect(client, SIGNAL(readyRead()), this, SLOT(receiveTcpData()));
    connect(client, SIGNAL(disconnected()), this, SLOT(clientDisonnected()));
    handClient[handSide] = client;

    qDebug() << "WhichFingers (" << (handSide == RIGHT_HAND ? "right" : "left") <<  "hand," << client->peerAddress() << client->peerPort() << ") connected !";

    bool bothHandsConnected = true;
    for (unsigned int i = 0; i < sizeof(handClient) / sizeof(QTcpSocket*); ++i) {
        if (handClient[i] == NULL) {
            bothHandsConnected = false;
            break;
        }
    }

    if (bothHandsConnected) {
        emit handsConnected();
    }
}

/*
 * Receive the response (ie the max values of each piezo) to a request
 */
void WhichFingers::receiveTcpData() {
    for (unsigned int i = 0; i < sizeof(handClient) / sizeof(QTcpSocket*); ++i) {
        QTcpSocket* client = handClient[i];

        if (client != NULL && client->bytesAvailable() >= (int) (sizeof(finger_response))) {
            qDebug() << "response for hand" << i << "in" << QDateTime::currentMSecsSinceEpoch() - lastRequestTimestamp << "ms";

            client->read((char*) &lastResponse, sizeof(finger_response));

            for (int i = 0; i < fingersToIdentify.size(); ++i) {
                finger_to_identify* fti = &fingersToIdentify[i];
                QHash<QHostAddress, finger_response>* responses = &(fti->responses);

                if (fti->packetId == lastResponse.id && !responses->contains(client->peerAddress())) {
                    responses->insert(client->peerAddress(), lastResponse);
                    if (fingersToIdentify[i].nbResponseToReceive == fingersToIdentify[i].responses.size()) {
                        identifyFinger(fingersToIdentify.at(i));
                        fingersToIdentify.removeAt(i);
                    }
                    break;
                }

            }
        }
    }

}

/*
 * Identify a finger by analyzing the values of the piezos
 */
void WhichFingers::identifyFinger(finger_to_identify fingerToId) {
    int piezoMax = 0;
    int piezoMaxVal = 0;
    QHashIterator<QHostAddress, finger_response> it(fingerToId.responses);

    while (it.hasNext()) {
        it.next();
        unsigned char handSide = hostHandSide[it.key()];

        for (unsigned int i = 0; i < sizeof(it.value().maxValues) / sizeof(char); ++i) {
            unsigned char val = it.value().maxValues[i];
            unsigned char fingerId = i + (handSide == RIGHT_HAND ? 5 : 0);

            if (val > piezoMaxVal && !fingers[fingerId].isPressed()) {
                piezoMax = fingerId;
                piezoMaxVal = val;
            }
        }
    }

    fingers[piezoMax].update(true, fingerToId.finger.getDevice(), fingerToId.finger.getTouchId());

    emit fingerIdentified(piezoMax);
}

/*
 *
 */
void WhichFingers::receiveUdpData() {
    while (udpServer->bytesAvailable() >= (signed) (sizeof(handshake_packet)) && udpServer->hasPendingDatagrams()) {
        QHostAddress sender;
        quint16 senderPort;
        handshake_packet packet;

        udpServer->readDatagram((char*) &packet, sizeof(handshake_packet), &sender, &senderPort);

        qDebug() << "WhichFingers (" << (packet.handSide == RIGHT_HAND ? "right" : "left") << "hand," << sender << senderPort << " ) is trying to establish a connection...";

        // Accept the connection by responding with the same packet (TODO : Refuse the connection if there is already one connected ?)
        hostHandSide[sender] = packet.handSide;
        udpSocket->writeDatagram((char*) &packet, sizeof(handshake_packet), sender, senderPort);
        udpSocket->flush();
    }
}

void WhichFingers::clientDisonnected() {
    for (unsigned int i = 0; i < sizeof(handClient) / sizeof(QTcpSocket*); ++i) {
        QTcpSocket* client = handClient[i];
        if (client != NULL && client->state() != QTcpSocket::ConnectedState) {
            qDebug() << "WhichFingers (" << (i == 0 ? "left" : "right") << ") disconnected";
            handClient[i] = NULL;
        }
    }

}

void WhichFingers::sendCalibrationRequest(int hand, unsigned char finger) {
    finger_request request;
    request.id = finger; // Finger id should be relative to the hand
    request.is_calibration = 1;


    QTcpSocket* client = handClient[hand];

    if (client != NULL) {
        if (client->write((char*) &request, sizeof(finger_request)) > 0) {
            client->flush();
        }
    } else {
        qDebug() << "Hardware prototype not connected";
    }
}

/*
 * Send a calibration request
 *
 * Knowing the finger that was used, the board will center its time window on the peak.
 * If there is two hands, then the calibration should be done with at least one finger of each hand
 */
void WhichFingers::calibrateOnPress(unsigned char fingerUsed) {
    int handSide = fingerUsed >= 5;
    sendCalibrationRequest(handSide, fingerUsed - (handSide ? 5 : 0));
}

/*
 * Reset all the calibration that was done on the board
 * This function should be called before doing a calibration using calibrateOnPress to clear
 * any previous calibration
 */
void WhichFingers::resetCalibration() {
    sendCalibrationRequest(0, 255);
    sendCalibrationRequest(1, 255);
}

/*
 * Send a request to every prototypes connected to the server to retrieve
 * the max values of each fingers
 *
 * Returns the number of request sent (this number should correspond to the number of prototype connected)
 */
int WhichFingers::sendRequest(long long int timestamp, unsigned char packetId) {
    int nbRequest = 0;
    finger_request request;
    request.id = packetId;
    request.is_calibration = 0;

    lastRequestTimestamp = timestamp;
    for (unsigned int i = 0; i < sizeof(handClient) / sizeof(QTcpSocket*); ++i) {
        QTcpSocket* client = handClient[i];

        if (client != NULL) {
            if (client->write((char*) &request, sizeof(finger_request)) > 0) {
                ++nbRequest;
                client->flush();
            }

        } else {
            qDebug() << "Hardware prototype not connected";
        }
    }

    return nbRequest;
}

void WhichFingers::devicePressed(int deviceId, int touchId) {
    finger_to_identify fingerToId;
    fingerToId.finger.update(true, deviceId, touchId);
    fingerToId.packetId = curPacketId;
    fingerToId.nbResponseToReceive = sendRequest(QDateTime::currentMSecsSinceEpoch(), curPacketId);
    curPacketId = (curPacketId + 1) % 255;

    if (fingerToId.nbResponseToReceive > 0) {
        fingersToIdentify.append(fingerToId);
    }
}

void WhichFingers::deviceReleased(int deviceId, int touchId) {
    for (unsigned int i = 0; i < sizeof(fingers) / sizeof(Finger); ++i) {
        Finger* finger = fingers + i;
        if (finger->isPressed() && finger->getDevice() == deviceId && finger->getTouchId() == touchId) {
            finger->update(false, -1, -1);
        }
    }
}

Finger* WhichFingers::getFingers() {
    return fingers;
}

WhichFingers::~WhichFingers() {
    for (unsigned int i = 0; i < sizeof(handClient) / sizeof(QTcpSocket*); ++i) {
        if (handClient[i] != NULL && handClient[i]->isOpen()) {
            handClient[i]->close();
        }
    }

    tcpServer->close();
    udpServer->close();
}
